/****************************************************************************
** Copyright (C) 1992-2000 Trolltech AS.  All rights reserved.
**
** This file is part of an example program for Qt.  This example
** program may be used, distributed and modified without limitation.
**
*****************************************************************************/

#include "life.h"

#include <qpainter.h>
#include <qdrawutil.h>
#include <qcheckbox.h>
#include <qevent.h>
#include <qapplication.h>
#include "math.h"
#include "time.h"
#include "stdio.h"
#include "stdlib.h"

int generation;			  /* Current generation no. */
int n_generation;		  /* Generation for obtaining optimal solution */
int avg_n_generation;	  /* Average of n_generation */
int cur_best;		      /* Best individual */
FILE *galog, *solution;	  /* Output files */

struct genotype           /* Genotype (GT), a member of the population */
{
	int	gene[NVARS];	  /* A string of variables */
	double	fitness;	  /* GT's fitness */
	double	rfitness;	  /* Relative fitness */
	double	cfitness;	  /* Cumulative fitness */
};

struct genotype population[POPSIZE+1];		/* Population */
struct genotype newpopulation[POPSIZE+1];	/* New population */
											/* replaces the old population */

// The main game of life widget

LifeWidget::LifeWidget( int s, QWidget *parent, const char *name )
    : QFrame( parent, name )
{
    SCALE = s;

	repaint_on = true;

    maxi = maxj = 50;
    setMinimumSize( MINSIZE * SCALE + 2 * BORDER,
		   MINSIZE * SCALE + 2 * BORDER );
    setMaximumSize( MAXSIZE * SCALE + 2 * BORDER,
		   MAXSIZE * SCALE + 2 * BORDER );
    setSizeIncrement( SCALE, SCALE);

    clear();
    resize( maxi * SCALE + 2 * BORDER , maxj * SCALE + 2 * BORDER );

	rule[0] = 0;
	rule[1] = 0;
	rule[2] = 1;
	rule[3] = 2;
	rule[4] = 0;
	rule[5] = 0;
	rule[6] = 0;
	rule[7] = 1;
	rule[8] = 0;

}


void LifeWidget::clear()
{
    current = 0;
    for ( int t = 0; t < 2; t++ )
	for ( int i = 0; i < MAXSIZE + 2; i++ )
	    for ( int j = 0; j < MAXSIZE + 2; j++ )
		cells[t][i][j] = FALSE;

    repaint();
}


// We assume that the size will never be beyond the maximum size set
// this is not in general TRUE, but in practice it's good enough for
// this program

void LifeWidget::resizeEvent( QResizeEvent * e )
{
    maxi = (e->size().width()  - 2 * BORDER) / SCALE;
    maxj = (e->size().height() - 2 * BORDER) / SCALE;
}


void LifeWidget::setPoint( int i, int j )
{
    if ( i < 1 || i > maxi || j < 1 || j > maxi )
	return;
    cells[current][i][j] = TRUE;
    repaint( index2pos(i), index2pos(j), SCALE, SCALE, FALSE );
}


void LifeWidget::mouseHandle( const QPoint &pos )
{
    int i = pos2index( pos.x() );
    int j = pos2index( pos.y() );
    setPoint( i, j );
}


void LifeWidget::mouseMoveEvent( QMouseEvent *e )
{
    mouseHandle( e->pos() );
}


void LifeWidget::mousePressEvent( QMouseEvent *e )
{
    if ( e->button() == QMouseEvent::LeftButton )
	mouseHandle( e->pos() );
}


void LifeWidget::nextGeneration()
{


    for ( int i = 1; i <= MAXSIZE; i++ ) {
		for ( int j = 1; j <= MAXSIZE; j++ ) {
			int t = cells[current][i - 1][j - 1]
					+ cells[current][i - 1][j]
					+ cells[current][i - 1][j + 1]
					+ cells[current][i][j - 1]
					+ cells[current][i][j + 1]
					+ cells[current][i + 1][j - 1]
					+ cells[current][i + 1][j]
					+ cells[current][i + 1][j + 1];

			cells[!current][i][j] = ( rule[t] == 2 || (rule[t] == 1 && cells[current][i][j]) );
				
//			cells[!current][i][j] = ( t == 3 ||
//						  t == 2 && cells[current][i][j] );
		}
    }
    current = !current;
    if( repaint_on ) repaint( FALSE );		// repaint without erase
}


void LifeWidget::paintEvent( QPaintEvent * e )
{
    int starti = pos2index( e->rect().left() );
    int stopi  = pos2index( e->rect().right() );
    int startj = pos2index( e->rect().top() );
    int stopj  = pos2index( e->rect().bottom() );

    if (stopi > maxi)
	stopi = maxi;
    if (stopj > maxj)
	stopj = maxj;

    QPainter paint( this );
    for ( int i = starti; i <= stopi; i++ ) {
	for ( int j = startj; j <= stopj; j++ ) {
	    if ( cells[current][i][j] )
		qDrawShadePanel( &paint, index2pos( i ), index2pos( j ),
				 SCALE - 1, SCALE - 1, colorGroup() );
	    else if ( cells[!current][i][j] )
		paint.eraseRect( index2pos( i ), index2pos( j ),
				 SCALE - 1, SCALE - 1);
	}
    }
    drawFrame( &paint );
}

void LifeWidget::initialize(void)
{
	int i, j;

	for( int k=0; k<9; k++ )
		rule[k] = 0;
	rule[2] = 1;
	rule[3] = 2;

	for (i = 0; i < POPSIZE+1; i++) {
		population[i].fitness = 0.;
		population[i].rfitness = 0.;
		population[i].cfitness = 0.;
		for (j = 0; j < NVARS; j++) {
			population[i].gene[j] = rule[j];
		}
	}
}


/*************************************************************/
/* Evaluation function: This takes a user defined function.  */
/* Each time this is changed, the code has to be recompiled. */
/*************************************************************/

void LifeWidget::evaluate(void)
{
	int mem;
	int i,j,k;
	float temp_fitness;
	int diff_a, diff_b;
	float diff;
	int num_of_on_cells, num_of_cur;
	double alpha_a = 1.0;
	double alpha_b = 0.1;


	for (mem = 0; mem < POPSIZE; mem++) {
		for(i=0;i<NVARS;i++){
			rule[i] = population[mem].gene[i];
		}
	    for ( i = 0; i <= MAXSIZE + 1; i++ ) 
			for ( j = 0; j <= MAXSIZE + 1; j++ ) 
				cells[current][i][j] = init_cells[i][j];

		temp_fitness = 0;
		num_of_on_cells = 0;

		for( k = 0; k < MAX_GENERATION ; k++ ){
			nextGeneration();
			for ( i = 1; i <= MAXSIZE - 2 ; i++ ) {
					for ( j = 1; j <= MAXSIZE - 2; j++ ) {
						diff_a = 1*cells[current][i][j] + 2*cells[current][i+1][j] + 3*cells[current][i+2][j] + 
								 4*cells[current][i][j+1] + 5*cells[current][i+1][j+1] + 6*cells[current][i+2][j+1] + 
								 7*cells[current][i][j+2] + 8*cells[current][i+1][j+2] + 9*cells[current][i+2][j+2];
						diff_b = 1*cells[!current][i][j] + 2*cells[!current][i+1][j] + 3*cells[!current][i+2][j] + 
								 4*cells[!current][i][j+1] + 5*cells[!current][i+1][j+1] + 6*cells[!current][i+2][j+1] + 
								 7*cells[!current][i][j+2] + 8*cells[!current][i+1][j+2] + 9*cells[!current][i+2][j+2];
						diff = abs(diff_a - diff_b);
						if( diff < MAX_FIT ) temp_fitness += diff/MAX_FIT;
						else temp_fitness += (MAX_DIFF - diff)/MAX_FIT;
					}
			}
			num_of_cur = 0;
			for ( i = 1; i <= MAXSIZE ; i++ ) {
				for ( j = 1; j <= MAXSIZE ; j++ ) {
						if( cells[!current][i][j] )	num_of_cur++;
				}
			}

			num_of_on_cells += (MAXSIZE*MAXSIZE - abs(num_of_init - num_of_cur));
					
			

		}


		population[mem].fitness = alpha_a*(float)(temp_fitness/MAXSIZE/MAXSIZE/MAX_GENERATION) + alpha_b*(float)(num_of_on_cells/MAXSIZE/MAXSIZE/MAX_GENERATION);



	}
	

}

float Diff( bool * a, bool * b )
{
	return 1;
}


/***************************************************************/
/* Keep_the_best function: This function keeps track of the    */
/* best member of the population. Note that the last entry in  */
/* the array Population holds a copy of the best individual    */
/***************************************************************/

void LifeWidget::keep_the_best()
{
	int mem;
	int i;
	cur_best = 0;       /* stores the index of the best individual */

	for (mem = 0; mem < POPSIZE; mem++) {
		if (population[mem].fitness > population[POPSIZE].fitness) {
			cur_best = mem;
			population[POPSIZE].fitness = population[mem].fitness;
		}
	}
	/* Once the best member in the population is found, copy the genes */
	for (i = 0; i < NVARS; i++)
		population[POPSIZE].gene[i] = population[cur_best].gene[i];
}

/****************************************************************/
/* Elitist function: The best member of the previous generation */
/* is stored as the last in the array. If the best member of    */
/* the current generation is worse than the best member of the  */
/* previous generation, the latter one would replace the worst  */
/* member of the current population                             */
/****************************************************************/

void LifeWidget::elitist()
{
	int i;
	double best, worst;			/* Best and worst fitness values */
	int best_mem, worst_mem;	/* Indexes of the best and worst member */

	best = population[0].fitness;
	worst = population[0].fitness;
	for (i = 0; i < POPSIZE-1; ++i) {
		if(population[i].fitness > population[i+1].fitness) {      
			if (population[i].fitness >= best) {
				best = population[i].fitness;
				best_mem = i;
			}
			if (population[i+1].fitness <= worst) {
				worst = population[i+1].fitness;
				worst_mem = i + 1;
			}
		} else {
			if (population[i].fitness <= worst) {
				worst = population[i].fitness;
				worst_mem = i;
			}
			if (population[i+1].fitness >= best)
			{
				best = population[i+1].fitness;
				best_mem = i + 1;
			}
		}
	}
	/* If best individual from the new population is better than */
	/* the best individual from the previous population, then    */
	/* copy the best from the new population; else replace the   */
	/* worst individual from the current population with the     */
	/* best one from the previous generation                     */

	if (best > population[POPSIZE].fitness) n_generation = generation+1;
	if (best >= population[POPSIZE].fitness) {
		for (i = 0; i < NVARS; i++)
			population[POPSIZE].gene[i] = population[best_mem].gene[i];
		population[POPSIZE].fitness = population[best_mem].fitness;
	} else {
		for (i = 0; i < NVARS; i++)
			population[worst_mem].gene[i] = population[POPSIZE].gene[i];
		population[worst_mem].fitness = population[POPSIZE].fitness;
	} 
}
/**************************************************************/
/* Selection function: Standard proportional selection for    */
/* maximization problems incorporating elitist model - makes  */
/* sure that the best member survives                         */
/**************************************************************/

void LifeWidget::Select(void)
{
	int mem, i, j;
	double sum = 0;
	double p;

	/* Find total fitness of the population */
	for (mem = 0; mem < POPSIZE; mem++) {
		sum += population[mem].fitness;
	}

	/* Calculate relative fitness */
	for (mem = 0; mem < POPSIZE; mem++) {
		population[mem].rfitness =  population[mem].fitness/sum;
	}
	population[0].cfitness = population[0].rfitness;

	/* Calculate cumulative fitness */
	for (mem = 1; mem < POPSIZE; mem++) {
		population[mem].cfitness =  population[mem-1].cfitness +       
			population[mem].rfitness;
	}

	/* Finally select survivors using cumulative fitness. */
	for (i = 0; i < POPSIZE; i++) { 
		p = (double)rand() /RAND_MAX;
		if (p < population[0].cfitness)
			newpopulation[i] = population[0];      
		else {
			for (j = 0; j < POPSIZE;j++)      
				if (p >= population[j].cfitness && 
						p<population[j+1].cfitness)
					newpopulation[i] = population[j+1];
		}
	}

	/* Once a new population is created, copy it back */
	for (i = 0; i < POPSIZE; i++)
		population[i] = newpopulation[i];      
}

/***************************************************************/
/* Crossover selection: selects two parents that take part in  */
/* the crossover. Implements a single point crossover.          */
/***************************************************************/

void LifeWidget::crossover(void)
{
	int mem, one;
	int first  =  0;	/* Count of the number of members chosen */
	double x;

	for (mem = 0; mem < POPSIZE; ++mem) {
		x = (double)rand() / RAND_MAX;
		if (x < PXOVER) {
			++first;
			if (first % 2 == 0)
				Xover(one, mem);
			else
				one = mem;
		}
	}
}
/**************************************************************/
/* Crossover: performs crossover of the two selected parents. */
/**************************************************************/

void LifeWidget::Xover(int one, int two)
{
	int i;
	int point;                	/* Crossover point */

	/* Select crossover point */
	if(NVARS > 1) {
		if(NVARS == 2)
			point = 1;
		else
			point = (rand() % (NVARS-1)) + 1;

		for (i = 0; i < point; i++)
			swap(&population[one].gene[i], &population[two].gene[i]);

	}
}

/*************************************************************/
/* Swap: A swap procedure that helps in swapping 2 variables */
/*************************************************************/

void LifeWidget::swap(int *x, int *y)
{
	int temp;

	temp = *x;
	*x = *y;
	*y = temp;
}

/**************************************************************/
/* Mutation: Random uniform mutation.                         */
/**************************************************************/

void LifeWidget::mutate(void)
{
	int i, j;
	double x,y;

	for (i = 0; i < POPSIZE; i++)
		for (j = 0; j < NVARS; j++) {
			x = (((double)rand() /RAND_MAX)-0.5)*6;
			y = (double)rand() /RAND_MAX;

			if (y < PMUTATION) {
				population[i].gene[j] += (int)x;
			}
			if( population[i].gene[j] > 2 || population[i].gene[j] < 0 )
				int djflskdfj = 0;
//			if (population[i].gene[j] > 19 ) population[i].gene[j] -= 19;
//			if (population[i].gene[j] < 0 ) population[i].gene[j] += 19;
			if (population[i].gene[j] > 2 ) population[i].gene[j] -= 2;
			if (population[i].gene[j] < 0 ) population[i].gene[j] += 2;
		}
}



/***************************************************************/
/* Report function: Reports progress of the simulation. Data   */
/* dumped into the  output file are separated by commas.       */
/***************************************************************/

void LifeWidget::report(void)
{
	int i;
	double best_val;		/* Best population fitness */
	double avg_fitness;	    /* Avg population fitness */
	double stddev;			/* Std. deviation of population fitness */
	double sum_square;		/* Sum of square for std. calc */
	double square_sum;		/* Square of sum for std. calc */
	double sum;				/* Total population fitness */

	sum = 0.0;
	sum_square = 0.0;

	for (i = 0; i < POPSIZE; i++)
	{
		sum += population[i].fitness;
		sum_square += population[i].fitness * population[i].fitness;
	}

	avg_fitness = sum/(double)POPSIZE;
	square_sum = avg_fitness * avg_fitness * POPSIZE;
	stddev = sqrt((sum_square - square_sum)/(POPSIZE - 1));
	best_val = population[POPSIZE].fitness;

	for (i = 0; i < NVARS; i++) {
		fprintf (galog,"%d ",population[POPSIZE].gene[i]);
	}

	fprintf(galog, "%5d       %10.9lf   %10.9lf   %10.9lf \n", generation, 
			best_val, avg_fitness, stddev);

}

/**************************************************************/
/* Main function: Each generation involves selecting the best */
/* members, performing crossover & mutation and then          */
/* evaluating the resulting population, until the terminating */
/* condition is satisfied                                     */
/**************************************************************/

void LifeWidget::gamain(void)
{
	int i,j, n;


	if ((galog = fopen("galog.txt","w"))==NULL) {
		exit(1);
	}
	if ((solution = fopen("solution.txt","w"))==NULL) {
		exit(1);
	}

	repaint_on = true;
	num_of_init = 0;
    for ( i = 0; i <= MAXSIZE + 1; i++ ) 
		for ( j = 0; j <= MAXSIZE + 1; j++ ) {
			init_cells[i][j] = cells[current][i][j];
			if( init_cells[i][j] ) num_of_init++;
		}

	avg_n_generation = 0;
	for(n=0; n<NRUNS; n++) {
		generation = 0;
		n_generation = 0;

		if(n==0){
			fprintf(galog, "\n	Rules			Generation    Best        Average       Standard \n");
			fprintf(galog,   "	[0] - [8]		number        value       fitness       deviation \n");
		}

		srand( (unsigned)time(NULL) );
		initialize();
		evaluate();
	
		keep_the_best();
		while(generation<MAXGENS) {
			generation++;
			Select();
			crossover();
			mutate();
			report();
			evaluate();
			elitist();
		}

		if( population[POPSIZE].fitness >= OPTIMAL ) {
			avg_n_generation += n_generation;
		} else {
			avg_n_generation += MAXGENS;
			n_generation = MAXGENS;
		}
		fprintf(solution,"Best member of %d-th run : \n", n+1);

		for (i = 0; i < NVARS; i++) {
			fprintf (solution,"%d ",population[POPSIZE].gene[i]);
		}
		fprintf(solution,"Best fitness = %10.9lf",population[POPSIZE].fitness);
		fprintf(solution,"  for %d generation\n\n",n_generation);
		printf("%d-th run completed\n", n+1);
	}
	fprintf(solution, "Average Generation for Obtaining Optimal Solution = %lf\n",
				(double)avg_n_generation/NRUNS);
	fclose(galog);
	fclose(solution);

	repaint_on = true;

}


